@@ -1,6 +1,6 @@
-/* Copyright (c) 2015-2017, 2019-2021 The Khronos Group Inc.
- * Copyright (c) 2015-2017, 2019-2021 Valve Corporation
- * Copyright (c) 2015-2017, 2019-2021 LunarG, Inc.
+/* Copyright (c) 2015-2017, 2019-2022 The Khronos Group Inc.
+ * Copyright (c) 2015-2017, 2019-2022 Valve Corporation
+ * Copyright (c) 2015-2017, 2019-2022 LunarG, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -26,9 +26,92 @@
 #include <cassert>
 #include <limits>
 #include <memory>
+#include <map>
 #include <unordered_map>
-#include <unordered_set>
+#include <set>
 #include <algorithm>
+#include <iterator>
+#include <type_traits>
+
+#ifdef USE_ROBIN_HOOD_HASHING
+#include "robin_hood.h"
+#else
+#include <unordered_set>
+#endif
+
+// namespace aliases to allow map and set implementations to easily be swapped out
+namespace layer_data {
+
+#ifdef USE_ROBIN_HOOD_HASHING
+template <typename T>
+using hash = robin_hood::hash<T>;
+
+template <typename Key, typename Hash = robin_hood::hash<Key>, typename KeyEqual = std::equal_to<Key>>
+using unordered_set = robin_hood::unordered_set<Key, Hash, KeyEqual>;
+
+template <typename Key, typename T, typename Hash = robin_hood::hash<Key>, typename KeyEqual = std::equal_to<Key>>
+using unordered_map = robin_hood::unordered_map<Key, T, Hash, KeyEqual>;
+
+template <typename Key, typename T>
+using map_entry = robin_hood::pair<Key, T>;
+
+// robin_hood-compatible insert_iterator (std:: uses the wrong insert method)
+template <typename T>
+class insert_iterator : public std::iterator<std::output_iterator_tag, void, void, void, void> {
+  public:
+    typedef typename T::value_type value_type;
+    typedef typename T::iterator iterator;
+    insert_iterator(T &t, iterator i) : container(&t), iter(i) {}
+
+    insert_iterator &operator=(const value_type &value) {
+        auto result = container->insert(value);
+        iter = result.first;
+        ++iter;
+        return *this;
+    }
+
+    insert_iterator &operator=(value_type &&value) {
+        auto result = container->insert(std::move(value));
+        iter = result.first;
+        ++iter;
+        return *this;
+    }
+
+    insert_iterator &operator*() { return *this; }
+
+    insert_iterator &operator++() { return *this; }
+
+    insert_iterator &operator++(int) { return *this; }
+
+  private:
+    T *container;
+    typename T::iterator iter;
+};
+#else
+template <typename T>
+using hash = std::hash<T>;
+
+template <typename Key, typename Hash = std::hash<Key>, typename KeyEqual = std::equal_to<Key>>
+using unordered_set = std::unordered_set<Key, Hash, KeyEqual>;
+
+template <typename Key, typename T, typename Hash = std::hash<Key>, typename KeyEqual = std::equal_to<Key>>
+using unordered_map = std::unordered_map<Key, T, Hash, KeyEqual>;
+
+template <typename Key, typename T>
+using map_entry = std::pair<Key, T>;
+
+template <typename T>
+using insert_iterator = std::insert_iterator<T>;
+#endif
+
+// Can't alias variadic functions even on C++14, and wrapping the 14 implementation wouldn't gain us anything...
+// leave it in unconditionally and for backwards compatibility
+template<typename T, typename... Args>
+constexpr std::unique_ptr<T> make_unique(Args&&... args) {
+    return std::unique_ptr<T>(new T(std::forward<Args>(args)...));
+}
+
+}  // namespace layer_data
 
 // A vector class with "small string optimization" -- meaning that the class contains a fixed working store for N elements.
 // Useful in in situations where the needed size is unknown, but the typical size is known  If size increases beyond the
@@ -83,6 +166,8 @@ class small_vector {
         other.size_ = 0;
     }
 
+    ~small_vector() { clear(); }
+
     bool operator==(const small_vector &rhs) const {
         if (size_ != rhs.size_) return false;
         auto value = begin();
@@ -184,7 +269,15 @@ class small_vector {
         return GetWorkingStore()[pos];
     }
 
-    // Like std::vector::back, calling back on an empty container causes undefined behavior
+    // Like std::vector:: calling front or back on an empty container causes undefined behavior
+    reference front() {
+        assert(size_ > 0);
+        return GetWorkingStore()[0];
+    }
+    const_reference front() const {
+        assert(size_ > 0);
+        return GetWorkingStore()[0];
+    }
     reference back() {
         assert(size_ > 0);
         return GetWorkingStore()[size_ - 1];
@@ -274,6 +367,8 @@ class small_container {
 
     inner_container_type inner_cont;
 
+    value_type_helper helper;
+
   public:
     small_container() {
         for (int i = 0; i < N; ++i) {
@@ -431,7 +526,7 @@ class small_container {
 
     bool contains(const Key &key) const {
         for (int i = 0; i < N; ++i) {
-            if (small_data_allocated[i] && value_type_helper().compare_equal(small_data[i], key)) {
+            if (small_data_allocated[i] && helper.compare_equal(small_data[i], key)) {
                 return true;
             }
         }
@@ -446,7 +541,7 @@ class small_container {
 
     std::pair<iterator, bool> insert(const value_type &value) {
         for (int i = 0; i < N; ++i) {
-            if (small_data_allocated[i] && value_type_helper().compare_equal(small_data[i], value)) {
+            if (small_data_allocated[i] && helper.compare_equal(small_data[i], value)) {
                 iterator it;
                 it.parent = this;
                 it.index = i;
@@ -454,7 +549,7 @@ class small_container {
             }
         }
         // check size() first to avoid hashing key unnecessarily.
-        auto iter = inner_cont.size() > 0 ? inner_cont.find(value_type_helper().get_key(value)) : inner_cont.end();
+        auto iter = inner_cont.size() > 0 ? inner_cont.find(helper.get_key(value)) : inner_cont.end();
         if (iter != inner_cont.end()) {
             iterator it;
             it.parent = this;
@@ -465,7 +560,7 @@ class small_container {
             for (int i = 0; i < N; ++i) {
                 if (!small_data_allocated[i]) {
                     small_data_allocated[i] = true;
-                    value_type_helper().assign(small_data[i], value);
+                    helper.assign(small_data[i], value);
                     iterator it;
                     it.parent = this;
                     it.index = i;
@@ -483,7 +578,7 @@ class small_container {
 
     typename inner_container_type::size_type erase(const Key &key) {
         for (int i = 0; i < N; ++i) {
-            if (small_data_allocated[i] && value_type_helper().compare_equal(small_data[i], key)) {
+            if (small_data_allocated[i] && helper.compare_equal(small_data[i], key)) {
                 small_data_allocated[i] = false;
                 return 1;
             }
@@ -520,15 +615,16 @@ class small_container {
 
 // Helper function objects to compare/assign/get keys in small_unordered_set/map.
 // This helps to abstract away whether value_type is a Key or a pair<Key, T>.
-template <typename Key, typename T>
+template <typename MapType>
 class value_type_helper_map {
-  public:
-    bool compare_equal(const std::pair<const Key, T> &lhs, const Key &rhs) const { return lhs.first == rhs; }
-    bool compare_equal(const std::pair<const Key, T> &lhs, const std::pair<const Key, T> &rhs) const {
-        return lhs.first == rhs.first;
-    }
+    using PairType = typename MapType::value_type;
+    using Key = typename std::remove_const<typename PairType::first_type>::type;
 
-    void assign(std::pair<const Key, T> &lhs, const std::pair<Key, T> &rhs) const {
+  public:
+    bool compare_equal(const PairType &lhs, const Key &rhs) const { return lhs.first == rhs; }
+    bool compare_equal(const PairType &lhs, const PairType &rhs) const { return lhs.first == rhs.first; }
+
+    void assign(PairType &lhs, const PairType &rhs) const {
         // While the const_cast may be unsatisfactory, we are using small_data as
         // stand-in for placement new and a small-block allocator, so the const_cast
         // is minimal, contained, valid, and allows operators * and -> to avoid copies
@@ -536,7 +632,7 @@ class value_type_helper_map {
         lhs.second = rhs.second;
     }
 
-    Key get_key(const std::pair<const Key, T> &value) const { return value.first; }
+    Key get_key(const PairType &value) const { return value.first; }
 };
 
 template <typename Key>
@@ -551,11 +647,12 @@ class value_type_helper_set {
 
 template <typename Key, typename T, int N = 1>
 class small_unordered_map
-    : public small_container<Key, std::pair<const Key, T>, std::unordered_map<Key, T>, value_type_helper_map<Key, T>, N> {
+    : public small_container<Key, typename layer_data::unordered_map<Key, T>::value_type, layer_data::unordered_map<Key, T>,
+                             value_type_helper_map<layer_data::unordered_map<Key, T>>, N> {
   public:
     T &operator[](const Key &key) {
         for (int i = 0; i < N; ++i) {
-            if (this->small_data_allocated[i] && value_type_helper_map<Key, T>().compare_equal(this->small_data[i], key)) {
+            if (this->small_data_allocated[i] && this->helper.compare_equal(this->small_data[i], key)) {
                 return this->small_data[i].second;
             }
         }
@@ -566,7 +663,7 @@ class small_unordered_map
             for (int i = 0; i < N; ++i) {
                 if (!this->small_data_allocated[i]) {
                     this->small_data_allocated[i] = true;
-                    value_type_helper_map<Key, T>().assign(this->small_data[i], std::make_pair(key, T()));
+                    this->helper.assign(this->small_data[i], {key, T()});
 
                     return this->small_data[i].second;
                 }
@@ -577,7 +674,7 @@ class small_unordered_map
 };
 
 template <typename Key, int N = 1>
-class small_unordered_set : public small_container<Key, Key, std::unordered_set<Key>, value_type_helper_set<Key>, N> {};
+class small_unordered_set : public small_container<Key, Key, layer_data::unordered_set<Key>, value_type_helper_set<Key>, N> {};
 
 // For the given data key, look up the layer_data instance from given layer_data_map
 template <typename DATA_T>
@@ -602,10 +699,8 @@ void FreeLayerDataPtr(void *data_key, small_unordered_map<void *, DATA_T *, 2> &
 template <typename DATA_T>
 DATA_T *GetLayerDataPtr(void *data_key, std::unordered_map<void *, DATA_T *> &layer_data_map) {
     DATA_T *debug_data;
-    typename std::unordered_map<void *, DATA_T *>::const_iterator got;
-
     /* TODO: We probably should lock here, or have caller lock */
-    got = layer_data_map.find(data_key);
+    auto got = layer_data_map.find(data_key);
 
     if (got == layer_data_map.end()) {
         debug_data = new DATA_T;
@@ -626,22 +721,32 @@ void FreeLayerDataPtr(void *data_key, std::unordered_map<void *, DATA_T *> &laye
     layer_data_map.erase(got);
 }
 
+namespace layer_data {
+
+struct in_place_t {};
+static constexpr in_place_t in_place{};
+
 // A C++11 approximation of std::optional
 template <typename T>
-struct Optional {
+class optional {
   protected:
     union Store {
         Store(){};   // Do nothing.  That's the point.
-        ~Store(){};  // Not safe to destroy this object outside of it's stateful contain to clean up T if any.
+        ~Store(){};  // Not safe to destroy this object outside of its stateful container to clean up T if any.
         typename std::aligned_storage<sizeof(T), alignof(T)>::type backing;
         T obj;
     };
 
   public:
-    Optional() : init_(false) {}
-    ~Optional() {
-        if (init_) store_.obj.~T();
-    }
+    optional() : init_(false) {}
+
+    template <typename... Args>
+    explicit optional(in_place_t, const Args &...args) { emplace(args...); }
+    optional(const optional &other) : init_(false) { *this = other; }
+    optional(optional &&other) : init_(false) { *this = std::move(other); }
+
+    ~optional() { DeInit(); }
+
     template <typename... Args>
     T &emplace(const Args &...args) {
         init_ = true;
@@ -665,9 +770,98 @@ struct Optional {
         return nullptr;
     }
     operator bool() const { return init_; }
+    bool has_value() const { return init_; }
 
+    optional &operator=(const optional &other) {
+        if (other.has_value()) {
+            if (has_value()) {
+                store_.obj = other.store_.obj;
+            } else {
+                emplace(other.store_.obj);
+            }
+        } else {
+            DeInit();
+        }
+        return *this;
+    }
+
+    optional &operator=(optional &&other) {
+        if (other.has_value()) {
+            if (has_value()) {
+                store_.obj = std::move(other.store_.obj);
+            } else {
+                emplace(std::move(other.store_.obj));
+            }
+        } else {
+            DeInit();
+        }
+        return *this;
+    }
+
+    T& operator*() & {
+        assert(init_);
+        return store_.obj;
+    }
+    const T& operator*() const& {
+        assert(init_);
+        return store_.obj;
+    }
+    T&& operator*() && {
+        assert(init_);
+        return std::move(store_.obj);
+    }
+    const T&& operator*() const&& {
+        assert(init_);
+        return std::move(store_.obj);
+    }
   protected:
+    inline void DeInit() {
+        if (init_) {
+            store_.obj.~T();
+            init_ = false;
+        }
+    }
     Store store_;
     bool init_;
 };
+
+// Partial implementation of std::span for C++11
+template <typename T>
+class span {
+  public:
+    using pointer = T *;
+    using iterator = pointer;
+    using const_iterator = T const *;
+
+    span() = default;
+    span(pointer start, size_t n) : data_(start), count_(n) {}
+    template <typename Iterator>
+    span(Iterator start, Iterator end) : data_(&(*start)), count_(end - start) {}
+    template <typename Container>
+    span(const Container &c) : data_(c.data()), count_(c.size()) {}
+
+    iterator begin() { return data_; }
+    const_iterator begin() const { return data_; }
+
+    iterator end() { return data_ + count_; }
+    const_iterator end() const { return data_ + count_; }
+
+    T &operator[](int i) { return data_[i]; }
+    const T &operator[](int i) const { return data_[i]; }
+
+    T &front() { return *data_; }
+    const T &front() const { return *data_; }
+
+    T &back() { return *(data_ + (count_ - 1)); }
+    const T &back() const { return *(data_ + (count_ - 1)); }
+
+    size_t size() const { return count_; }
+
+    pointer data() { return data_; }
+
+  private:
+    pointer data_ = {};
+    size_t count_ = 0;
+};
+}  // namespace layer_data
 #endif  // LAYER_DATA_H
